﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace EtkezoFilozofusok
{

  //az asztal osztálya
  class Asztal
  {
    int hanySzemelyes; // hány személyes az asztal
    Semaphore[] villak; // a villákat reprezentáló szemaforok

    //konstruktor, param: hány személyes legyen az asztal
    public Asztal(int hanySzemelyes)
    {
      //példány változók inicializálása
      this.hanySzemelyes = hanySzemelyes;

      //dinamikus változók lefoglalása
      villak = new Semaphore[hanySzemelyes];
      for (int i = 0; i < hanySzemelyes; ++i)
        villak[i] = new Semaphore(1, 1); //bináris szemoforok, kezdetben mindegyik szabad
    }

    //ha valaki felveszi a villát
    public void villakFel(int ki)
    {
      Monitor.Enter(this); //megfogom a monitort, boztosítva a kölcsönös kizárást
      bool nincsMegKetVilla = false; // még nincs két villánk
      do
      {
        nincsMegKetVilla = villak[ki].WaitOne(0, true); //szabad az egyik villám?
        if (nincsMegKetVilla) // ha rögtön fel tud venni egyet
        {
          nincsMegKetVilla = villak[(ki + 1) % hanySzemelyes].WaitOne(0, true); //szabad a másik villám?
          if (!nincsMegKetVilla) //ha nem tudtam felvenni a másodikat
            villak[ki].Release(); //elengedem azt, amit fel tudtam venni
        }
        if (!nincsMegKetVilla) //ha nincs meg mindkét villa
        {
          Monitor.Wait(this); //várakozom, esélyt adva a többieknek, hogy letegyék a villákat (*)
        }
      } while (!nincsMegKetVilla); //amig nem sikerült felvenni a villákat, nem lépek ki
      Monitor.Exit(this); //elhagyom a monitort
    }

    //leteszi valaki a villáit
    public void villakLe(int ki)
    {
      Monitor.Enter(this); //kölcsönös kizárás a kritikus szakaszra
      villak[ki].Release(); //elengedem az egyik,...
      villak[(ki + 1) % hanySzemelyes].Release();   //... majd a másik villámat
      Monitor.PulseAll(this); //jelzem, hogy letettem a villákat -> (*)
      Monitor.Exit(this); //elhagyom a monitort
    }
  }

  //A filozófus osztálya
  class Filosz
  {
    int sorszam; //hányadik filozófus
    Asztal asztal; //melyik asztalnál ül

    //konstruktor, param: sorszáma, és hogy melyik asztalnál ül
    public Filosz(int sorszam, Asztal asztal)
    {
      //példány változók inicializálása
      this.asztal = asztal;
      this.sorszam = sorszam;
    }

    //A filozófus futási metódusa, ez fog külön szálban futni
    public void run()
    {
      while (true) //a végtelenségig csinálja a következőket:
      {
        asztal.villakFel(sorszam); //felveszi a villákat
        Console.WriteLine(sorszam + ". eszik...");  //szól, hogy eszik
        Thread.Sleep(5000); //ilyen sokáig eszik
        asztal.villakLe(sorszam); //leteszi a villákat
        Console.WriteLine(sorszam + ". gondolkodik..."); //elgondolkodik
        Thread.Sleep(5000);  //ilyen sokáig gondolkodik
      }
    }
  }

  //A főprogram osztálya
  class Program
  {

    //A főprogram
    static void Main()
    {
      //kell egy asztal, ami jelen esetben 5 személyes
      Asztal asztal = new Asztal(5);

      //kell 5 filozófus, a megfelelő sorszámokkal, és ugyanannál az asztalnál ülnek
      Filosz[] filosz = {
        new Filosz(0, asztal),
        new Filosz(1, asztal),
        new Filosz(2, asztal),
        new Filosz(3, asztal),
        new Filosz(4, asztal)
      };

      //A szálak elindításához kellenek
      ThreadStart[] threadStart = new ThreadStart[filosz.Length];

      //kiírjuk az instrukcókat
      Console.WriteLine("ENTER - kilépés");

      //A szálak lefoglalása
      Thread[] filoszok = new Thread[filosz.Length];
      for (int i = 0; i < filoszok.Length; ++i) //minden filozófusra:
      {
        threadStart[i] = new ThreadStart(filosz[i].run); //a saját futási fügvényével fut
        filoszok[i] = new Thread(threadStart[i]); //létrehozzuk a saját szálált
        filoszok[i].Start(); //elindítjuk a szálat
      }

      //Várunk egy ENTER-re
      Console.ReadLine();

      //Leállítjuk a szálakat
      for (int i = 0; i < filoszok.Length; ++i) //minden filozófusra...
      {
        filoszok[i].Abort();
      }
    }
  }


}
